#include "AMDiS.h"

using namespace std;




static double epsNorm1(const WorldVector x, const double eps)
{
  static double result = 0.0;
  static int k;
  
  result = 0.0;
  for(k=0; k<x.getSize(); k++){
    result +=  x[k]*x[k];
  }
  result += eps*eps;
  result = sqrt( result );
  TEST_EXIT( result > 1.e-10)( "eps_norm = %e \n", result);
  result = 1.0/result;  

  return(result);
}


static double epsNorm(const WorldVector x, const double eps)
{
  static double result = 0.0;
  static int  k;
  
  result = 0.0;

  for(k=0; k<x.getSize(); k++){
    result +=  x[k]*x[k];
  }
  result += eps*eps;
  result = sqrt( result );

  return(result);
}

static void epsNormalize(const WorldVector x, const double eps, WorldVector y)
{
  static int k;
  
  for(k=0; k<x.getSize(); k++)
    y[k] = x[k] * epsNorm1(x,eps);
  
  return;
}



class Circle : public AbstractFunction<double, double>
{
public:
  Circle(double radius_) : radius(radius_) {};

  const double& f(const double& alpha) const
  {
    return radius;
  };

protected:
  double radius;
};


class PerturbedCircle : public AbstractFunction<double, double>
{
public:
  PerturbedCircle(double radius_, double strength_, int period_) : 
    radius(radius_), strength(strength_), period(period_) {};

  const double& f(const double &alpha) const
  {
    static double result; 
    result = 1.0 + strength * cos( period * alpha );
    result *= radius;
    return result;
  };

protected:
  double radius;
  double strength;
  int period;
};


/****************************************************************************/
/*       signed distance function for levelset 0 circle with radius         */
/****************************************************************************/


static double signedDistStar(const WorldVector& x, const WorldVector& midPoint, 
			     AbstractFunction<double, double> *radius, double eps)
{
  WorldVector  x_trans;
  double norm_x;
  double result;
  double alpha;
  
  int k;
  for(k=0; k < x.getSize(); k++) {
    x_trans[k] = x[k] - midPoint[k];
  }

  //x_trans = (WorldVector&) x - (WorldVector&) mid_point;
  
  norm_x = epsNorm(x_trans, eps);
  TEST_EXIT( norm_x > 1.e-15)( "eps_norm = %e \n", norm_x);
  
  if( x_trans[0] >= 0.0 )
    alpha = acos( x_trans[0] / norm_x );
  else 
    alpha = 2.0*M_PI - acos( x_trans[0] / norm_x );
  
  result = ( norm_x - (*radius)(alpha)); 
    
  return(result);
};



/****************************************************************************/
/* anisotropy function and derivative                                        */
/****************************************************************************/

class Gamma : public AbstractFunction<double, WorldVector>
{
public:
  Gamma(double alpha_) : alpha(alpha_) {};

  const double& f(const WorldVector& x) const
  {
    static double result;
    int  k;

    result = alpha * x[0] * x[0] + 1.e-10;
    for( k=1; k<x.getSize(); k++)
      result += x[k]*x[k];

    result = sqrt(result);
    return result;
  };

protected:
  double alpha;
};




class Gamma_z : public AbstractFunction<WorldVector, WorldVector>
{
public:
  Gamma_z(double alpha_) : alpha(alpha_) {};

  const WorldVector& f(const WorldVector& x) const
  {
    static WorldVector result;
    int  k;
    double norm_1 = 0.0;
    static Gamma gamma(alpha);
    double norm = gamma(x);

    TEST_EXIT( norm  > 1.e-15)( "gamma(z) = %e \n", norm);
    norm_1 = 1.0/norm;

    result[0] =  alpha * x[0] * norm_1;

    for( k=1; k<x.getSize(); k++)
      result[k] = x[k] * norm_1;

    return result;
  };

protected:
  double alpha;
};


/****************************************************************************/
/*     Abstract functions used by operators                                 */
/****************************************************************************/

/*      function for second order operator A     */

class ProjectionGradNorm : public AbstractFunction<WorldMatrix, WorldVector>
{
public:
  ProjectionGradNorm( double epsilon_) : epsilon(epsilon_) {};

  const WorldMatrix& f(const WorldVector& x) const
  {
    static WorldMatrix result;
    static WorldVector gradNorm;
    static double      norm;
    int k,l;

    norm  = epsNorm(x, epsilon);
    epsNormalize(x, epsilon,  gradNorm);
 
    for(k = 0; k<x.getSize(); k++)
      for(l = 0; l<= k; l++){
	if( k == l )  
	  result[k][l] = 1.0 - gradNorm[k]*gradNorm[l];
	else
	  result[k][l] =  -gradNorm[k]*gradNorm[l];
	result[k][l] *= norm;
	result[l][k] = result[k][l];
      }

    return result;
  };

protected:
  double alpha;
  double epsilon;
};


/*      function for second order operator B     */

class GammaNorm : public AbstractFunction<double, WorldVector>
{
public:
  GammaNorm( double epsilon_, double alpha_) : epsilon(epsilon_), alpha(alpha_) {};

  const double& f(const WorldVector& x) const
  {
    static double result;
    static WorldVector gradNorm;
    static double      norm1;
    static Gamma gamma(alpha);
    int k,l;

    norm1  = epsNorm1(x, epsilon);
    epsNormalize(x, epsilon,  gradNorm);

    result = gamma(gradNorm) * norm1;
    return result;
  };

protected:
  double epsilon;
  double alpha;
};





/*        function for rhs F    */
class GammaGradNorm : public AbstractFunction<WorldVector, WorldVector>
{
public:
  GammaGradNorm(double alpha_, double epsilon_) : alpha(alpha_), epsilon(epsilon_) {};

  const WorldVector& f(const WorldVector& x) const
  {
    static WorldVector result;
    static WorldVector gradNorm;
    static Gamma gamma(alpha);

    epsNormalize(x, epsilon,  gradNorm);
    result = gamma( gradNorm ) *  gradNorm;

    return result;
  };

protected:
  double alpha;
  double epsilon;
};


/*        function for rhs G   */
class Gamma_zGradNorm : public AbstractFunction<WorldVector, WorldVector>
{
public:
  Gamma_zGradNorm(double alpha_, double epsilon_) : alpha(alpha_), epsilon(epsilon_) {};

  const WorldVector& f(const WorldVector& x) const
  {
    static WorldVector result;
    static WorldVector gradNorm;
    static Gamma_z gamma_z(alpha);

    epsNormalize(x, epsilon,  gradNorm);
    result = gamma_z( gradNorm );

    return result;
  };

protected:
  double alpha;
  double epsilon;
};


/**********************************************/
/* create stationary problem                  */
/**********************************************/

class LevelsetSpace : public Problem<double>
{
public:
  LevelsetSpace() 
    : Problem<double>("levelset->space")
  {
    GET_PARAMETER(1, name + "->lambda", "%f", &lambda);

    A   = NEW DOFMatrix(feSpace, "A");
    B   = NEW DOFMatrix(feSpace, "B");
    M   = NEW DOFMatrix(feSpace, "M");
    M_lump   = NEW DOFMatrix(feSpace, "M_lump");
    M_1 = NEW DOFMatrix(feSpace, "M_1");

    f   = NEW DOFVector<double>(feSpace, "f");
    g   = NEW DOFVector<double>(feSpace, "g");
    rhs = NEW DOFVector<double>(feSpace, "rhs");
  };

  ~LevelsetSpace() 
  {
    DELETE A;
    DELETE B;
    DELETE M;
    DELETE M_lump;
    DELETE M_1;
    DELETE f;
    DELETE g;
  };

  DOFMatrix *getMatrixA() { return A; };
  DOFMatrix *getMatrixB() { return B; };
  DOFMatrix *getMatrixM() { return M; };
  DOFVector<double> *getVectorF() { return f; };
  DOFVector<double> *getVectorG() { return g; };

  static int buildAfterCoarsenFunction(ElInfo *elInfo) 
  {
    const char *bound;

    if(traversePtr->useGetBound)
      bound = traversePtr->feSpace->getBasisFcts()->getBound(elInfo, NULL);
    else
      bound = NULL;
  
    traversePtr->A->assemble(1.0, elInfo, bound);
    traversePtr->B->assemble(1.0, elInfo, bound);
    traversePtr->M->assemble(1.0, elInfo, bound);
    traversePtr->f->assemble(1.0, elInfo, bound);
    traversePtr->g->assemble(1.0, elInfo, bound);

    return 0;
  };

  void buildAfterCoarsen(Flag flag)
  {
    FUNCNAME("LevelsetSpace::buildAfterCoarsen");

    MSG("%d DOFs for %s\n", feSpace->getAdmin()->getUsedSize(), feSpace->getName().c_str());

    mesh->dofCompress();

    Flag assembleFlag = 
      flag | A->getAssembleFlag() | B->getAssembleFlag() | M->getAssembleFlag() | 
      f->getAssembleFlag() | g->getAssembleFlag();

    if(useGetBound)
      assembleFlag |= Mesh::FILL_BOUND;

    A->clear();
    B->clear();
    M->clear();
    f->set(0.0);
    g->set(0.0);
   
  
    traversePtr = this;

    mesh->traverse(-1, assembleFlag, &buildAfterCoarsenFunction);
    //dirichletBound(gFct, fh, uh, NULL);           /*  boundary values         */
  };
  
  void solve() {};

  double estimate() { return 0.0; };

private:

#if 0

void matrixLump(DOF_MATRIX *matrix, DOF_MATRIX *matrix_diag)
{
  int k,i,icol;
  MATRIX_ROW *row;

  FUNCNAME("matrix_lump");
  clear_dof_matrix(matrix_diag);
  FOR_ALL_DOFS(gd->fe_space->admin,    
	       matrix_diag->matrix_row[dof] = get_matrix_row(gd->fe_space);
	       matrix_diag->matrix_row[dof]->entry[0]=0.0;
	       row = matrix->matrix_row[dof];
	       while( row ) {
		 for(i=0; i<ROW_LENGTH; i++){
		   icol = row->col[i];
		   if( ENTRY_USED(icol) )
		     matrix_diag->matrix_row[dof]->entry[0] +=
		       matrix->matrix_row[dof]->entry[i];
		   else{
		     if (icol == NO_MORE_ENTRIES)
		       break;
		   }
		 }
		 row = row->next;
	       }
	       matrix_diag->matrix_row[dof]->col[0] = dof;
	       matrix_diag->matrix_row[dof]->col[1] = NO_MORE_ENTRIES;
	       for( k = 2; k < ROW_LENGTH; k++)
	       matrix_diag->matrix_row[dof]->col[k] = UNUSED_ENTRY;

	       matrix_diag->matrix_row[dof]->next = nil;
	       )
 return;
}

#endif

/****************************************************************************/
/*invertDiagMatrix(): inverts a positive definit diagonal matrix          */
/****************************************************************************/

void invertDiagMatrix(DOFMatrix *matrix, DOFMatrix *matrix_1)
{
  int k;
  FUNCNAME("invert_matrix_lump");
  matrix_1->clear();

  DOFMatrix::Iterator matrixIt(matrix, USED_DOFS);
  double entry = 0.0;
  DegreeOfFreedom dof = 0;

  for(matrixIt.reset(); !matrixIt.end(); ++matrixIt) {
    dof = matrixIt.getPosition();
    if( (*matrix)[dof][0].entry > 1.e-10 ){
      entry = 1.0 / (*matrix)[dof][0].entry;
      matrix_1->addSparseDofEntry(1.0, dof, dof, entry);
    }
    else 
      ERROR_EXIT("matrix[%d][%d] = %f", dof, dof, (*matrix)[dof][0].entry );

  }
   

 return;
}





private:
  DOFMatrix *A, *B, *M, *M_lump, *M_1;
  DOFVector<double> *rhs, *f, *g;
  
  double *timestepPtr;

  double lambda;

  static LevelsetSpace *traversePtr;

  friend class Levelset;
};

LevelsetSpace *LevelsetSpace::traversePtr = NULL;

/**********************************************/
/* create instationary problem                */
/**********************************************/

class Levelset : public ProblemInstat<double>
{
public:
  Levelset(LevelsetSpace *levelsetSpace)
    : ProblemInstat<double>("levelset", (Problem<double>*) levelsetSpace)
  {
    
    timestep = adaptInstat->getTimestepPtr();
    levelsetSpace->timestepPtr = timestep;

    int problemNumber = -1;
    GET_PARAMETER(1, name + "->problem number","%d", &problemNumber);
    epsilon = 0.1;
    GET_PARAMETER(1, name + "->epsilon","%f", &epsilon);
    alpha = 1.0;
    GET_PARAMETER(1, name + "->alpha","%f", &alpha);


    switch(problemNumber) {
    case 0:
      initialFunction = new InitialFunction0(epsilon);
      break;

    case 1:
      initialFunction = new InitialFunction1(epsilon);
      break;

    case -1:
      ERROR_EXIT("no problem number specified\n");
      break;
    default: ERROR_EXIT("non existing problem number %d\n", problemNumber);
    }

  };


  // === timestep ===================================================
  void initTimestep() 
  { 
    uhOld->copy(*(problemStat->getUh())); 
  };

  void closeTimestep() 
  {
    FUNCNAME("closeTimestep");  
  };

  // === initial functions ==================================
  void solveInitial() 
  {
    //problemStat->getMesh()->dofCompress();
    //problemStat->getBoundaryFunction()->setTime(0.0);
    //problemStat->getUh()->interpol(initialFunction);
  };

  void setEvaluationTimes() {};
 
  double *getTimestepPtr() { return timestep; };


  double getAlpha(){ return alpha; };
  double getEpsilon(){ return epsilon; };


private:
  AbstractFunction<double, WorldVector> *initialFunction;

  double *timestep;
  double epsilon;
  double alpha;

private:
  class InitialFunction0 : public AbstractFunction<double, WorldVector> 
  {
  public:
    InitialFunction0(double epsilon_) : epsilon(epsilon_) {}

    const double& f(const WorldVector& x) const
    {
      static WorldVector midPoint(DEFAULT_VALUE, 2.0);
      static double result;
      static Circle circle1(1.0);
      result = signedDistStar( x, midPoint, &circle1, epsilon ); 
      
    };  

  private:
    double epsilon;

  };

  class InitialFunction1 : public AbstractFunction<double, WorldVector> 
  {
  public:
    InitialFunction1(double epsilon_) : epsilon(epsilon_) {}

    const double& f(const WorldVector& x) const
    {
      static WorldVector midPoint(DEFAULT_VALUE, 2.0);
      static double result;
      static PerturbedCircle perturbedCircle1(1.0, 0.3, 6);
      result = signedDistStar( x, midPoint, &perturbedCircle1, epsilon ); 
    };  

  private:
    double epsilon;
  };
 
};



 
int main(int argc, char** argv)
{
  TEST_EXIT(argc == 2)("usage: levelset initfile\n");

  Parameters::init(false, argv[1]);

  LevelsetSpace *levelsetSpace = new LevelsetSpace;
  
  Levelset *levelset = new Levelset(levelsetSpace);
  double epsilon = levelset->getEpsilon();
  double alpha   = levelset->getAlpha();


  // create  mass matrix M operator
  Operator *operatorM = NEW Operator(Operator::MATRIX_OPERATOR | 
				     Operator::VECTOR_OPERATOR,
				     levelsetSpace->getFESpace());

  operatorM->addZeroOrderTerm(NEW SimpleZeroOrderTerm);

  operatorM->setUhOld(levelset->getUhOld());

  levelsetSpace->addMatrixOperator(operatorM);
  levelsetSpace->addVectorOperator(operatorM);

  // create  mass matrix B operator
  Operator *operatorB = NEW Operator(Operator::MATRIX_OPERATOR,
				     levelsetSpace->getFESpace());

  operatorB->addSecondOrderTerm(NEW FctGradient_SOT(levelset->getUhOld(),
						    new GammaNorm(alpha, epsilon),
						    100));

  // create  mass matrix A operator
  Operator *operatorA = NEW Operator(Operator::MATRIX_OPERATOR,
				     levelsetSpace->getFESpace());

  operatorA->addSecondOrderTerm(NEW MatrixGradient_SOT(levelset->getUhOld(),
						    new ProjectionGradNorm(epsilon),
						    100));

  // create rhs F 
  Operator *operatorF = NEW Operator(Operator::VECTOR_OPERATOR,
				     levelsetSpace->getFESpace());

  operatorF->addFirstOrderTerm(NEW  VectorGradient_FOT(levelset->getUhOld(),
						       new Gamma_zGradNorm(alpha,epsilon),
						       100, false));


  // create rhs G 
  Operator *operatorG = NEW Operator(Operator::VECTOR_OPERATOR,
				     levelsetSpace->getFESpace());

  operatorG->addFirstOrderTerm(NEW  VectorGradient_FOT(levelset->getUhOld(),
						       new GammaGradNorm(alpha,epsilon),
						       100, false));


  levelsetSpace->getMatrixA()->addOperator(operatorA);
  levelsetSpace->getMatrixB()->addOperator(operatorB);
  levelsetSpace->getMatrixM()->addOperator(operatorM);
  levelsetSpace->getVectorF()->addOperator(operatorF);
  levelsetSpace->getVectorG()->addOperator(operatorG);


  levelsetSpace->addVectorOperator(operatorM);


  levelset->adaptMethodInstat();

  return 0;
}
