#include "NonLinUpdater.h"
#include <algorithm>

namespace AMDiS {

  template<typename VectorType>
  void NewtonS<VectorType>::init()
  {
    TEST_EXIT(this->vectorCreator)("no vectorCreator\n");
    b = this->vectorCreator->create();
    y = this->vectorCreator->create();
  }

  template<typename VectorType>
  void NewtonS<VectorType>::exit()
  {
    if(b) this->vectorCreator->free(b);
    if(y) this->vectorCreator->free(y);
  }

  template<typename VectorType>
  int NewtonS<VectorType>::nlsolve(MatVecMultiplier<VectorType> *matVec,
				   VectorType *x, VectorType *rhs,
				   Preconditioner<VectorType> *lPrecon,
				   Preconditioner<VectorType> *rPrecon)
  {
    FUNCNAME("NewtonS::nlsolve");

    *b = *rhs;

    //   // copy operators from fh to b
    //   std::vector<Operator*>::iterator op;
    //   std::vector<double*>::iterator fac;
    //   for(op = rhs->getOperatorsBegin(), 
    // 	fac = rhs->getOperatorFactorBegin();
    //       op != rhs->getOperatorsEnd();
    //       ++op, ++fac) 
    //   {
    //     b->addOperator(*op, *fac);
    //   }

    double err, err_old, tau;
    int    iter, j, n, m, mmax, halved;

    /*--------------------------------------------------------------------------*/
    /*--- Newton initialization ------------------------------------------------*/
    /*--------------------------------------------------------------------------*/

    this->nonLinUpdater->update(/*x, */false, b);          /*-- update F(x) ------------*/

    if(this->usedNorm == NO_NORM || this->usedNorm == L2_NORM) {
      err = err_old = L2Norm(b);
    } else {
      err = err_old = H1Norm(b);
    }

    this->initial_residual = err;

    INFO(this->info,2)("iter. |     residual |     red. |    n |  m |\n");
    INFO(this->info,2)("%5d | %12.5e | -------- | ---- | -- |\n", 0, err);

    if ((this->residual = err) < this->tolerance)
      {
	INFO(this->info,4)("finished succesfully\n");
	return(0);
      }

    /*--- still initalization part ---------------------------------------------*/
  
    mmax        = std::max(2,std::min(restart,32));
    m           = 0;
    tau         = 1.0;
    halved      = true;


    /*--- start iterations -----------------------------------------------------*/

    for (iter = 1; iter <= this->maxIter; iter++)
      {
	/*--- Assemble DF(x) -------------------------------------------------------*/
	this->nonLinUpdater->update(/*x, */true, NULL);
	/*--- Initial guess is zero ------------------------------------------------*/
	*rhs = 0.0;
	/*--- solve linear system --------------------------------------------------*/
	n = solveLinearSystem(matVec, b, rhs, lPrecon, rPrecon);

	/*--- look for step size ---------------------------------------------------*/
	if (!halved) {
	  m = std::max(m-1,0);
	  tau = tau < 0.5 ? 2.0*tau : 1.0;
	}

	// remember original x
	*y = *x;

	for (j = 0; j <= mmax; j++) {
	  /*--- aim: |F(u_k+\tau d)| \le (1-0.5\tau) |F(u)| --------------------------*/
	  /*-- y = x ------------------------*/
	  //*y = *x;
	  *x = *y;

	  /*-- y -= tau*rhs -------------------*/
	  //      axpy(-tau, *rhs, *y);
	  axpy(-tau, *rhs, *x);

	  /*-- update F(y) ------------------*/
	  //      nonLinUpdater->update(y, false, b);
	  this->nonLinUpdater->update(/*x, */false, b);

	  if(this->usedNorm == NO_NORM || this->usedNorm == L2_NORM) {
	    err = L2Norm(b);
	  } else {
	    err = H1Norm(b);
	  }
 
 
	  if (err <= (1.0 - 0.5*tau)*err_old) {
	    halved = j > 0;
	    break;
	  } else {
	    if (m == mmax) break;
	
	    m++;
	    tau *= 0.5;
	  }
	}

	//*x = *y;       /*-- x = y  (update x!) ------------*/

	if (err_old <= 0) {
	  INFO(this->info,2)("%5d | %12.5e | -------- | %4d | %2d |\n", 
			     iter, err, n, m);
	} else {
	  INFO(this->info,2)("%5d | %12.5e | %8.2e | %4d | %2d |\n", 
			     iter, err, err/err_old, n, m);
	}

	if ((this->residual = err) < this->tolerance && m == 0)
	  {
	    INFO(this->info,4)("finished successfully\n");
	    return(iter);
	  } 
	else if (iter > this->maxIter)
	  {
	    if (this->info < 2)
	      INFO(this->info,1)("iter. %d, residual: %12.5e\n", iter, err);
	    INFO(this->info,1)("tolerance %e not reached\n", this->tolerance);
	    return(iter);
	  }
	err_old = err;
      }
    return(iter);  /*--- statement never reached -----------------------------*/
  }

}
