#include "Mesh.h"
#include "Parametric.h"
#include "Quadrature.h"
#include "PartitionElementData.h"

namespace AMDiS {

  template<typename T>
  const T& ParallelError<T>::errUFct(const DimVec<double>& lambda)
  {
    WorldVector<double> x;
    //   if (el_parametric) {
    //     ERROR_EXIT("not yet\n");
    //   } else {
    elinfo->coordToWorld(lambda, &x);
    //   }
    return((*pU)(x));
  }

  template<typename T>
  const WorldVector<T>& ParallelError<T>::grdErrUFct(const DimVec<double>& lambda)
  {
    WorldVector<double> x;
    //   if (el_parametric) {
    //     ERROR_EXIT("not yet\n");
    //   } else {
    elinfo->coordToWorld(lambda, &x);
    //   }
    return((*pGrdU)(x));
  }


  template<typename T>
  int ParallelError<T>::maxErrAtQpFct(ElInfo* el_info)
  {
    int i;
    double err;
    const double *u_vec, *uh_vec;
    //   Parametric *parametric = el_info->getMesh()->getParametric();

    elinfo = el_info;
    //   if (parametric && parametric->initElement(el_info)) {
    //     el_parametric = parametric;
    //   } else {
    //     el_parametric = NULL;
    //   }

    u_vec = quadFast->getQuadrature()->fAtQp(errU, NULL);

    //uh_el = errUh->getLocalVector(el_info->getElement(), NULL);
    //uh_vec = quadFast->uhAtQp(uh_el, NULL);

    uh_vec = errUh->getVecAtQPs(el_info,
				NULL,
				quadFast,
				NULL);

    int numPoints = quadFast->getNumPoints();

    for (i = 0; i < numPoints; i++)
      {
	err = u_vec[i] > uh_vec[i] ? u_vec[i] - uh_vec[i] : uh_vec[i] - u_vec[i];
	maxErr = max(maxErr, err);
      }

    return;  
  }

  template<typename T>
  double ParallelError<T>::maxErrAtQp(const AbstractFunction<T, WorldVector<double> >& u,
				      const DOFVector<T>& uh,
				      const Quadrature* q)
  {
    FUNCNAME("ParallelError<T>::maxErrAtQp");
    const FiniteElemSpace *fe_space;

    if (!(pU = &u))
      {
	ERROR("no function u specified; doing nothing\n");
	return(-1.0);
      }
    if (!(errUh = &uh)  ||  !(fe_space = uh->getFESpace()))
      {
	ERROR("no discrete function or no fe_space for it; doing nothing\n");
	return(-1.0);
      }

    if (!(basFct = fe_space->getBasisFcts()))
      {
	ERROR("no basis functions at discrete solution ; doing nothing\n");
	return(-1.0);
      }

    int dim = fe_space->getMesh()->getDim();

    if (!q)
      q = Quadrature::provideQuadrature(dim, 
					2*fe_space->getBasisFcts()->getDegree()
					- 2);
    quadFast = FastQuadrature::provideFastQuadrature(basFct, *q, INIT_PHI);

    maxErr = 0.0;


    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(fe_space->getMesh(), -1, 
					 Mesh::FILL_COORDS |
					 Mesh::CALL_LEAF_EL);
    while(elInfo) {
      PartitionElementData *partitionData = dynamic_cast<PartitionElementData*>
	(elInfo->getElement()->getElementData(PARTITION_ED));
      TEST_EXIT(partitionData)("no partition data\n");
      if(partitionData && partitionData->getPartitionStatus() == IN) {
	maxErrAtQpFct(elInfo);
      }
      elInfo = stack.traverseNext(elInfo);
    }

    //   fe_space->getMesh()->traverse(-1, 
    // 				Mesh::FILL_COORDS |
    // 				Mesh::CALL_LEAF_EL, 
    // 				maxErrAtQpFct);

    return(maxErr);
  }

  template<typename T>
  int ParallelError<T>::H1ErrFct(ElInfo* el_info)
  {
    int i, j;
    double err, err_2, h1_err_el, norm_el, norm2, det, exact;
    //int dim = el_info->getMesh()->getDim();
    //const DimVec<WorldVector<double> > &Lambda = el_info->getGrdLambda();
    //const T *uh_el;
    const WorldVector<double> *grdu_vec, *grduh_vec;
    //   const Parametric *parametric = el_info->getMesh()->getParametric();

    elinfo = el_info;
    //   if (parametric && parametric->initElement(el_info)) {
    //     el_parametric = parametric;
    //   }
    //   else {
    //     el_parametric = NULL;
    //   }

    grdu_vec = quadFast->getQuadrature()->grdFAtQp(grdErrU, NULL);

    //uh_el = errUh->getLocalVector(el_info->getElement(), NULL);

    //   if (el_parametric) {
    //     el_parametric->grd_lambda(el_info, NULL, 1, NULL, &Lambda, &det);
    //   }
    //   else {
    det = el_info->getDet();
    //   }

    //grduh_vec = quadFast->grdUhAtQp(Lambda, uh_el, NULL);
    grduh_vec = errUh->getGrdAtQPs(elinfo, NULL, quadFast, NULL);

    int numPoints = quadFast->getNumPoints();
    int dow = Global::getGeo(WORLD);

    for (h1_err_el = i = 0; i < numPoints; i++)
      {
	for (err_2 = j = 0; j < dow; j++) 
	  {
	    err = grdu_vec[i][j] - grduh_vec[i][j];
	    err_2 += sqr(err);
	  }
	h1_err_el += quadFast->getWeight(i)*err_2;
      }

    exact = det*h1_err_el;
    h1Err2 += exact;
    maxErr = max(maxErr, exact);

    if(writeInLeafData)
      el_info->getElement()->setEstimation(exact, component);

    if (relative)
      {
	for (norm_el = i = 0; i < numPoints; i++)
	  {
	    for (norm2 = j = 0; j < dow; j++) 
	      norm2 += sqr(grdu_vec[i][j]);
	    norm_el += quadFast->getWeight(i)*norm2;
	  }
	h1Norm2 += det*norm_el;
      }

    return 0;
  }

  template<typename T>
  double ParallelError<T>::H1Err(
				 const AbstractFunction<WorldVector<T>, WorldVector<double> >& grdU,
				 const DOFVector<T>&                                     uh,
				 //const Quadrature*                                       quad,
				 int                                                     relErr,
				 double*                                                 max,
				 bool writeLeafData,
				 int comp)
  {
    FUNCNAME("ParallelError<T>::H1Err");
    const FiniteElemSpace *fe_space;
  
    writeInLeafData = writeLeafData;

    component = comp;

    Quadrature *q = NULL;

    pGrdU = &grdU;

    errUh = &uh;

    if(!(fe_space = uh.getFESpace()))
      {
	ERROR("no fe_space for uh; doing nothing\n");
	return(0.0);
      }

    if (!(basFct = fe_space->getBasisFcts()))
      {
	ERROR("no basis functions at discrete solution ; doing nothing\n");
	return(0.0);
      }

    int dim = fe_space->getMesh()->getDim();
    int deg = grdU.getDegree();
    int degree = deg ? deg : 2 * fe_space->getBasisFcts()->getDegree() - 2;

    //if (!quad)
    q = Quadrature::provideQuadrature(dim, degree);
    quadFast = FastQuadrature::provideFastQuadrature(basFct, 
						     *q, 
						     INIT_GRD_PHI);

    relative = relErr;

    maxErr = h1Err2 = h1Norm2 = 0.0;


    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(fe_space->getMesh(), -1, 
					 Mesh::FILL_COORDS |
					 Mesh::CALL_LEAF_EL |
					 Mesh::FILL_DET | 
					 Mesh::FILL_GRD_LAMBDA);
    while(elInfo) {
      PartitionElementData *partitionData = dynamic_cast<PartitionElementData*>
	(elInfo->getElement()->getElementData(PARTITION_ED));
      TEST_EXIT(partitionData)("no partition data\n");
      if(partitionData && partitionData->getPartitionStatus() == IN) {
	H1ErrFct(elInfo);
      }
      elInfo = stack.traverseNext(elInfo);
    }

    //   fe_space->getMesh()->traverse(-1, 
    // 				Mesh::FILL_COORDS |
    // 				Mesh::CALL_LEAF_EL |
    // 				Mesh::FILL_DET | 
    // 				Mesh::FILL_GRD_LAMBDA, 
    // 				H1ErrFct);

    if (relative) {
      relNorm2 = h1Norm2+1.e-15;

      TraverseStack stack;
      ElInfo *elInfo = stack.traverseFirst(fe_space->getMesh(), -1, 
					   Mesh::CALL_LEAF_EL);
      while(elInfo) {
	PartitionElementData *partitionData = dynamic_cast<PartitionElementData*>
	  (elInfo->getElement()->getElementData(PARTITION_ED));
	TEST_EXIT(partitionData)("no partition data\n");
	if(partitionData && partitionData->getPartitionStatus() == IN) {
	  relFct(elInfo);
	}
	elInfo = stack.traverseNext(elInfo);
      }

      //     fe_space->getMesh()->traverse(-1, 
      // 				  Mesh::CALL_LEAF_EL, 
      // 				  relFct);

      h1Err2 /= relNorm2;
      maxErr /= relNorm2;
    }

    if (max) {
      double totalMax;
      MPI::COMM_WORLD.Allreduce(&maxErr, &totalMax, 1, MPI_DOUBLE, MPI_MAX);
      *max = totalMax;
    }

    double totalH1Err2;
    MPI::COMM_WORLD.Allreduce(&h1Err2, &totalH1Err2, 1, MPI_DOUBLE, MPI_SUM);

    return(sqrt(totalH1Err2));
  }

  template<typename T>
  int ParallelError<T>::L2ErrFct(ElInfo* el_info)
  {
    int i;
    double err, det, l2_err_el, norm_el, exact;
    const double *u_vec, *uh_vec;
    //   Parametric *parametric = const_cast<Parametric*>(el_info->getMesh()->getParametric());

    elinfo = el_info;

    //   if (parametric && parametric->initElement(el_info)) {
    //     el_parametric = parametric;
    //   }
    //   else {
    //     el_parametric = NULL;
    //   }

    u_vec = quadFast->getQuadrature()->fAtQp(errU, NULL);

    //uh_el = errUh->getLocalVector(el_info->getElement(), NULL);
    //uh_vec = quadFast->uhAtQp(uh_el, NULL);

    uh_vec = errUh->getVecAtQPs(el_info,
				NULL,
				quadFast,
				NULL);

    //   if (el_parametric) {
    //     el_parametric->det(el_info, NULL, 1, NULL, &det);
    //   }
    //   else {
    //     det = el_info->calcDet();
    //   }
 
    det = el_info->getDet();
  
    int numPoints = quadFast->getNumPoints();
  
    for (l2_err_el = i = 0; i < numPoints; i++)
      {
	err = u_vec[i] - uh_vec[i];
	l2_err_el += quadFast->getWeight(i)*sqr(err);
      }

    exact = det*l2_err_el;
    l2Err2 += exact;
    maxErr = max(maxErr, exact);

    if(writeInLeafData) {
      el_info->getElement()->setEstimation(exact, component);
    }

    if (relative)
      {
	for (norm_el = i = 0; i < numPoints; i++)
	  norm_el += quadFast->getWeight(i)*sqr(u_vec[i]);
	l2Norm2 += det*norm_el;
      }

    return 0;
  }

  template<typename T>
  double ParallelError<T>::L2Err(const AbstractFunction<T, WorldVector<double> >& u,
				 const DOFVector<T>& uh,
				 //Quadrature* quad,
				 int relErr,
				 double* max,
				 bool writeLeafData,
				 int comp)
  {
    FUNCNAME("ParallelError<T>::L2Err");
    const FiniteElemSpace *fe_space;

    Quadrature *q = NULL;

    writeInLeafData = writeLeafData;
  
    component = comp;

    if (!(pU = &u))
      {
	ERROR("no function u specified; doing nothing\n");
	return(0.0);
      }
    if (!(errUh = &uh)  ||  !(fe_space = uh.getFESpace()))
      {
	ERROR("no discrete function or no fe_space for it; doing nothing\n");
	return(0.0);
      }

    if (!(basFct = fe_space->getBasisFcts()))
      {
	ERROR("no basis functions at discrete solution ; doing nothing\n");
	return(0.0);
      }

    int dim = fe_space->getMesh()->getDim();
    int deg = u.getDegree();
    int degree = deg ? deg :  2 * fe_space->getBasisFcts()->getDegree() - 2;

    //if (!quad)
    q = Quadrature::provideQuadrature(dim, degree);
    quadFast = FastQuadrature::provideFastQuadrature(basFct, *q, INIT_PHI);

    relative = relErr;

    maxErr = l2Err2 = l2Norm2 = 0.0;

    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(fe_space->getMesh(), -1, 
					 Mesh::FILL_COORDS |
					 Mesh::FILL_DET |
					 Mesh::FILL_GRD_LAMBDA |
					 Mesh::CALL_LEAF_EL);
    while(elInfo) {
      PartitionElementData *partitionData = dynamic_cast<PartitionElementData*>
	(elInfo->getElement()->getElementData(PARTITION_ED));
      TEST_EXIT(partitionData)("no partition data\n");
      if(partitionData && partitionData->getPartitionStatus() == IN) {
	L2ErrFct(elInfo);
      }
      elInfo = stack.traverseNext(elInfo);
    }

    //   fe_space->getMesh()->traverse(-1, 
    // 				Mesh::FILL_COORDS |
    // 				Mesh::FILL_DET |
    // 				Mesh::FILL_GRD_LAMBDA |
    // 				Mesh::CALL_LEAF_EL,
    // 				L2ErrFct);

    if (relative) {
      relNorm2 = l2Norm2+1.e-15;
      elInfo = stack.traverseFirst(fe_space->getMesh(), -1, 
				   Mesh::CALL_LEAF_EL);
      while(elInfo) {
	PartitionElementData *partitionData = dynamic_cast<PartitionElementData*>
	  (elInfo->getElement()->getElementData(PARTITION_ED));
	TEST_EXIT(partitionData)("no partition data\n");
	if(partitionData && partitionData->getPartitionStatus() == IN) {
	  relFct(elInfo);
	}
	elInfo = stack.traverseNext(elInfo);
      }

      //     fe_space->getMesh()->traverse(-1, Mesh::CALL_LEAF_EL, relFct);
      l2Err2 /= relNorm2;
    }

    if (max) {
      double totalMax;
      MPI::COMM_WORLD.Allreduce(&maxErr, &totalMax, 1, MPI_DOUBLE, MPI_MAX);
      *max = totalMax;
    }

    double totalL2Err2;
    MPI::COMM_WORLD.Allreduce(&l2Err2, &totalL2Err2, 1, MPI_DOUBLE, MPI_SUM);
    
    return(sqrt(totalL2Err2));
  }

  template<typename T>
  int ParallelError<T>::relFct(ElInfo* elinfo)
  {
    double exact = elinfo->getElement()->getEstimation(component);
    exact /= relNorm2;
    if(writeInLeafData)
      elinfo->getElement()->setEstimation(exact, component);
    return 0;
  }

}
