// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ParallelProblem.h */

#ifndef AMDIS_PARALLELPROBLEM_H
#define AMDIS_PARALLELPROBLEM_H

#include <map>
#include <vector>
#include <set>

#include "ProblemTimeInterface.h"
#include "ProblemIterationInterface.h"
#include "AdaptInfo.h"
#include "mpi.h"

namespace AMDiS {

  /// Interface for parallel problems
  class ParallelProblemInterface
  {
  public:
    virtual ~ParallelProblemInterface() {};
    virtual void initParallelization(AdaptInfo *adaptInfo) = 0;
    virtual void exitParallelization(AdaptInfo *adaptInfo) = 0;
  };

  /// Schablonen Klasse
  class ParallelProblemBase : public ParallelProblemInterface,
			      public ProblemIterationInterface,
			      public ProblemTimeInterface
  {
  public:
    ParallelProblemBase(std::string name,
			ProblemIterationInterface *iterationIF,
			ProblemTimeInterface *timeIF);

    virtual ~ParallelProblemBase() {}

    /** \brief
     * Must return true, if a new partitioning of the domain (due to unbalanced
     * calculation times) have to be done.
     */
    virtual bool doPartitioning(AdaptInfo *adaptInfo, double localWeightSum) = 0;

    virtual bool doBuildGlobalSolution(AdaptInfo *adaptInfo) = 0;

    /// Set for each element on the partitioning level the number of leaf elements.
    virtual double setElemWeights(AdaptInfo *adaptInfo) = 0;

    virtual void partitionMesh(AdaptInfo *adaptInfo) = 0;

    virtual void refineOverlap(AdaptInfo *adaptInfo) = 0;    

    virtual void globalRefineOutOfPartition(AdaptInfo *adaptInfo) = 0;    

    virtual void createOverlap(AdaptInfo *adaptInfo) = 0;

    virtual void exchangeDOFVectors(AdaptInfo *adaptInfo) = 0;

    virtual void coarsenOutOfPartition(AdaptInfo *adaptInfo) = 0;

    virtual void synchronizeMeshes(AdaptInfo *adaptInfo) = 0;

    virtual void exchangeRankSolutions(AdaptInfo *adaptInfo) = 0;

    virtual void buildGlobalSolution(AdaptInfo *adaptInfo) = 0;

    virtual void exitParallelization(AdaptInfo *adaptInfo);

    virtual void setTime(AdaptInfo *adaptInfo) {
      if (timeIF_) 
	timeIF_->setTime(adaptInfo);
    }

    virtual void initTimestep(AdaptInfo *adaptInfo) {
      if (timeIF_) 
	timeIF_->initTimestep(adaptInfo);
    }

    virtual void closeTimestep(AdaptInfo *adaptInfo);

    virtual void solveInitialProblem(AdaptInfo *adaptInfo) {
      if (timeIF_)
	timeIF_->solveInitialProblem(adaptInfo);
    }
  
    virtual void transferInitialSolution(AdaptInfo *adaptInfo) {
      if (timeIF_) 
	timeIF_->transferInitialSolution(adaptInfo);
    }

    virtual void beginIteration(AdaptInfo *adaptInfo) {
      iterationIF_->beginIteration(adaptInfo);
    }

    virtual Flag oneIteration(AdaptInfo *adaptInfo, Flag toDo = FULL_ITERATION);

    virtual void endIteration(AdaptInfo *adaptInfo) {
      iterationIF_->endIteration(adaptInfo);
    }

    virtual void startDelayedTimestepCalculation() {}

    virtual bool existsDelayedCalculation() {
      return false;
    }

    MPI::Intracomm* getMpiComm() {
      return &mpiComm;
    }

    bool getDebugServerProcess() {
      return debugServerProcess;
    }

  protected:
    ProblemIterationInterface *iterationIF_;

    ProblemTimeInterface *timeIF_;

    clock_t computationStart;

    double partitioningTime;

    /// The rank of the current process.
    int mpiRank;

    /// Overall number of processes.
    int mpiSize;

    /** \brief
     * MPI communicator collected all processes, which should
     * be used for calculation. The Debug procces is not included
     * in this communicator.
     */
    MPI::Intracomm mpiComm;

    /// The MPI group, which is used for the communicator \ref mpiComm.
    MPI::Group amdisGroup;

    /// Defines the debug mode. If it is 1, a debug server will be started on rank 0.
    int debugMode;

    bool groupIsSet;
    /** \brief
     * If true, the current process is the debug server. Otherwise, the
     * processes is a working processes calculating part of the solution.
     */
    bool debugServerProcess;
  };

  // =========================================================================
  // ===== class ParallelProblem =============================================
  // =========================================================================

  class ParMetisPartitioner;
  class RefinementManager;
  class CoarseningManager;
  class ElInfo;
  class MeshStructure;

  class ParallelProblem : public ParallelProblemBase
  {
  public:
    ParallelProblem(std::string name,
		    ProblemIterationInterface *iterationIF,
		    ProblemTimeInterface *timeIF,
		    std::vector<DOFVector<double>*> vectors,
		    Mesh *mesh,
		    RefinementManager *rm,
		    CoarseningManager *cm);

    virtual ~ParallelProblem();

    virtual bool doPartitioning(AdaptInfo *adaptInfo, double localWeightSum);

    virtual bool doBuildGlobalSolution(AdaptInfo *adaptInfo);

    virtual void partitionMesh(AdaptInfo *adaptInfo);

    virtual void refineOverlap(AdaptInfo *adaptInfo);

    virtual void createOverlap(AdaptInfo *adaptInfo);    

    virtual void globalRefineOutOfPartition(AdaptInfo *adaptInfo);

    virtual void coarsenOutOfPartition(AdaptInfo *adaptInfo);

    virtual void synchronizeMeshes(AdaptInfo *adaptInfo);

    virtual double setElemWeights(AdaptInfo *adaptInfo);

    virtual void globalRefinements();

    void exchangeRankSolutions(AdaptInfo *adaptInfo,
			       Mesh *workMesh,
			       std::vector<DOFVector<double>*> rankSolutions);

    void exchangeDOFVector(AdaptInfo *adaptInfo,
			   DOFVector<double> *vec);

    void buildGlobalSolution(AdaptInfo *adaptInfo,
			     std::vector<DOFVector<double>*> rankSolutions,
			     DOFVector<double> *globalSolution);

    void createOverlap(int level, int overlap, bool openOverlap,
		       std::map<Element*, int> &overlapDistance);

    void fillVertexPartitions(int level, int overlap, bool openOverlap,
			      std::map<Element*, int> &overlapDistance);

    void setRepartitionSteps(int steps) { 
      repartitionSteps_ = steps; 
    }

    void puEveryTimestep(bool pu) { 
      puEveryTimestep_ = pu; 
    }

    void addDOFVector(DOFVector<double> *vec) {
      dofVectors_.push_back(vec);
    }

    /** \brief
     * Every process creates the mesh structure code of its mesh, and all
     * processes broadcast their mesh structure codes.
     */
    void exchangeMeshStructureCodes(MeshStructure *structures);

    static bool writeElement(ElInfo *elInfo);

    virtual void startDebugServer() {}

    virtual void serialize(std::ostream&) {}

    virtual void deserialize(std::istream&) {}

  protected:
    ///
    double errors2map(std::map<int, double> &errMap, int comp, bool add);

    ///
    std::vector<int> iList;

    ///
    std::string name_;

    /// Mesh of the problem.
    Mesh *mesh;

    ///
    RefinementManager *refinementManager;

    ///
    CoarseningManager *coarseningManager;

    /// Pointer to the paritioner which is used to devide a mesh into partitions.
    ParMetisPartitioner *partitioner;

    /** \brief
     * Stores to every coarse element index the number of the partition it 
     * corresponds to.
     */
    std::map<int, int> partitionVec;

    /** \brief
     * Stores an old partitioning of elements. To every element index the number
     * of the parition it corresponds to is stored.
     */
    std::map<int, int> oldPartitionVec;    

    /// Weights for the elements, i.e., the number of leaf elements within this element.
    std::map<int, double> elemWeights;

    /// Stores to every element the set of partitions it corresponds to.
    std::map<Element*, std::set<int> > elementPartitions_;

    /// Stores to every DOF the set of partitions it corresponds to.
    std::map<DegreeOfFreedom, std::set<int> > vertexPartitions;

    ///
    int repartitionSteps_;

    bool puEveryTimestep_;

    std::vector<DOFVector<double>*> dofVectors_;

    double upperPartThreshold_;

    double lowerPartThreshold_;

    int globalCoarseGridLevel_;

    int localCoarseGridLevel_;

    int globalRefinements_;

    std::map<Element*, int> overlapDistance_;

    int adaptiveThresholds_;

    double thresholdIncFactor_;

    double thresholdDecFactor_;

    double repartTimeFactor_;

    double minUpperTH_;

    double maxLowerTH_;
  };

  // =========================================================================
  // ===== class ParallelProblemScal =========================================
  // =========================================================================
  
  class FiniteElemSpace;
  class ProblemScal;
  class ProblemInstatScal;
  class Estimator;
  class Marker;

  class ParallelProblemScal : public ParallelProblem
  {
  public:
    ParallelProblemScal(std::string name,
			ProblemScal *problem,
			ProblemInstatScal *problemInstat,
			std::vector<DOFVector<double>*> vectors);

    virtual ~ParallelProblemScal();

    virtual void initParallelization(AdaptInfo *adaptInfo);

    virtual void exitParallelization(AdaptInfo *adaptInfo);

    virtual void exchangeDOFVectors(AdaptInfo *adaptInfo);

    virtual void exchangeRankSolutions(AdaptInfo *adaptInfo);

    virtual void buildGlobalSolution(AdaptInfo *adaptInfo);

    virtual void coarsenOutOfPartition(AdaptInfo *adaptInfo);

    virtual int getNumProblems();

    virtual ProblemStatBase *getProblem(int number = 0);

    virtual ProblemStatBase *getProblem(std::string name);

    void setEstimator(Estimator *est) 
    { 
      usersEstimator = est; 
    }

    void setMarker(Marker *marker) 
    { 
      usersMarker = marker; 
    }

    inline virtual std::string getName() 
    { 
      return name_; 
    }

  protected:
    ProblemScal *problem;

    ProblemInstatScal *problemInstat_;

    Estimator *oldEstimator;

    Marker *oldMarker;

    /// Vector of all process' solution DOFVectors.
    std::vector<DOFVector<double>*> rankSolution;

    ///
    Estimator *usersEstimator;

    ///
    Marker *usersMarker;
  };

  // =========================================================================
  // ===== class ParallelProblemVec ==========================================
  // =========================================================================
  
  class ProblemVec;
  class ProblemInstatVec;
  class SystemVector;

  class ParallelProblemVec : public ParallelProblem
  {
  public:
    ParallelProblemVec(std::string name,
		       ProblemVec *problem,
		       ProblemInstatVec *problemInstat,
		       std::vector<DOFVector<double>*> vectors);

    virtual ~ParallelProblemVec();

    virtual void initParallelization(AdaptInfo *adaptInfo);

    virtual void exitParallelization(AdaptInfo *adaptInfo);

    virtual void exchangeDOFVectors(AdaptInfo *adaptInfo);

    virtual void exchangeRankSolutions(AdaptInfo *adaptInfo);

    virtual void buildGlobalSolution(AdaptInfo *adaptInfo);

    virtual void coarsenOutOfPartition(AdaptInfo *adaptInfo);

    virtual int getNumProblems();

    virtual ProblemStatBase *getProblem(int number = 0);

    virtual ProblemStatBase *getProblem(std::string name);

    void setEstimator(std::vector<Estimator*> est) 
    { 
      usersEstimator = est; 
    }

    void setMarker(std::vector<Marker*> marker) 
    {
      usersMarker = marker; 
    }

    inline virtual std::string getName() 
    { 
      return name_; 
    }

    virtual void startDebugServer(AdaptInfo *adaptInfo);

    virtual void debugFunction(AdaptInfo *adaptInfo) {}

  protected:
    ///
    ProblemVec *problem;

    ///
    ProblemInstatVec *problemInstat_;

    ///
    std::vector<Estimator*> oldEstimator;

    ///
    std::vector<Marker*> oldMarker;

    /// Vector of all process' solution SystemVectors.
    std::vector<SystemVector*> rankSolution;

    ///
    std::vector<Estimator*> usersEstimator;

    ///
    std::vector<Marker*> usersMarker;

    /// Number of components of the vectorial problem.
    int nComponents;
  };

}

#endif
