// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ProblemNonLin.h */

#ifndef AMDIS_PROBLEMNONLIN_H_
#define AMDIS_PROBLEMNONLIN_H_

#include "ProblemScal.h"
#include "ProblemVec.h"
#include "NonLinUpdater.h"
#include "DOFVector.h"
#include "SystemVector.h"
#include "MatrixVector.h"

#if 0

namespace AMDiS {

  template<typename Vector> class NonLinSolver;

  // ==========================================================================
  // ===== class ProblemNonLinScal ============================================
  // ==========================================================================

  /**
   * \ingroup Problem
   * 
   * \brief
   * Standard implementation for a non linear problem.
   */
  class ProblemNonLinScal : public ProblemScal
  {
  public:
    /** \brief
     * Constructs a ProblemNonLin with given name.
     */
    ProblemNonLinScal(std::string name_)  
      : ProblemScal(name_.c_str()),
	u0_(NULL),
	nonLinSolver_(NULL),
	updater_(NULL)
    {}

    /** \brief
     * Destructor.
     */
    virtual ~ProblemNonLinScal() {}

    /** \brief
     * Initialisation of the problem.
     */
    virtual void initialize(Flag initFlag,
			    ProblemScal *adoptProblem = NULL,
			    Flag adoptFlag = INIT_NOTHING);

    /** \brief
     * Used in \ref initialize().
     */
    virtual void createUpdater() {
      updater_ = new NonLinUpdaterScal(systemMatrix);
    }

    /** \brief
     * Used in \ref initialize().
     */
    virtual void createNonLinSolver();

    /** \brief
     * Returns \ref nonLinSolver_.
     */
    inline NonLinSolver<DOFVector<double> >* getNonLinSolver() {
      return nonLinSolver_;
    }

    /** \brief
     * Returns \ref updater_.
     */
    inline NonLinUpdaterScal* getUpdater() {
      return updater_;
    }

    /** \brief
     * Sets \ref u0_ and interpolates it to \ref uh_.
     */
    inline void setU0(AbstractFunction<double, WorldVector<double> > *u0Fct) {
      u0_ = u0Fct;
      solution->interpol(u0_);
    }

    /** \brief
     * Sets \ref nonLinSolver_.
     */
    inline void setNonLinSolver(NonLinSolver<DOFVector<double> > *nonLinSolver) {
      nonLinSolver_=nonLinSolver;
    }

    /** \brief
     * Sets \ref updater.
     */
    inline void setUpdater(NonLinUpdaterScal *updater) {
      updater_ = updater;
    }

    /** \brief
     * Overrides ProblemScal::solve(). Uses the non linear solver 
     * \ref nonLinSolver_.
     */
    virtual void solve(AdaptInfo *adaptInfo);

    /** \brief
     * Overrides Problem::buildAfterCoarsen(). Sets dirichlet boundaries
     * in \ref A_ and \ref rhs_.
     */
    virtual void buildAfterCoarsen(AdaptInfo *adaptInfo, Flag);

  protected:
    /** \brief
     * Initial guess for the solution.
     */
    AbstractFunction<double, WorldVector<double> > *u0_;

    /** \brief
     * Non linear solver. Used in \ref solve().
     */
    NonLinSolver<DOFVector<double> > *nonLinSolver_;

    /** \brief
     * Updator
     */
    NonLinUpdaterScal *updater_;
  };

  // ==========================================================================
  // ===== class ProblemNonLinScal ============================================
  // ==========================================================================

  /**
   * \ingroup Problem
   * 
   * \brief
   * Standard implementation for a vector valued non linear problem.
   */
  class ProblemNonLinVec : public ProblemVec
  {
  public:
    /** \brief
     * Constructs a ProblemNonLin with given name.
     */
    ProblemNonLinVec(std::string name_)  
      : ProblemVec(name_.c_str()),
	nonLinSolver_(NULL),
	updater_(NULL)
    {
      u0_.resize(nComponents);
      u0_.set(NULL);
    }

    /** \brief
     * Sets \ref u0_ and interpolates it to \ref solution_.
     */
    inline void setU0(AbstractFunction<double, WorldVector<double> > *u0Fct,
		      int index) 
    {
      FUNCNAME("ProblemNonLinVec::setU0()");
      TEST_EXIT(index < nComponents)("invalid index\n");
      u0_[index] = u0Fct;
      solution->getDOFVector(index)->interpol(u0Fct);
    }

    /** \brief
     * Destructor.
     */
    virtual ~ProblemNonLinVec() {}

    /** \brief
     * Initialisation of the problem.
     */
    virtual void initialize(Flag initFlag,
			    ProblemVec *adoptProblem = NULL,
			    Flag adoptFlag = INIT_NOTHING);

    /** \brief
     * Used in \ref initialize().
     */
    virtual void createUpdater() {
      updater_ = new NonLinUpdaterVec(systemMatrix);
    }

    /** \brief
     * Used in \ref initialize().
     */
    virtual void createNonLinSolver();

    /** \brief
     * Overrides ProblemVec::solve(). Uses the non linear solver 
     * \ref nonLinSolver_.
     */
    virtual void solve(AdaptInfo *adaptInfo);

    /** \brief
     * Overrides Problem::buildAfterCoarsen(). Sets dirichlet boundaries
     * in \ref A_ and \ref rhs_.
     */
    virtual void buildAfterCoarsen(AdaptInfo *adaptInfo, Flag);

    /** \brief
     * Returns /ref nonLinSolver_.
     */
    inline NonLinSolver<SystemVector> *getNonLinSolver() {
      return nonLinSolver_;
    }

    /** \brief
     * Returns \ref updater.
     */
    inline NonLinUpdaterVec* getUpdater() {
      return updater_;
    }

    /** \brief
     * Sets \ref nonLinSolver_.
     */
    inline void setNonLinSolver(NonLinSolver<SystemVector> *nonLinSolver) {
      nonLinSolver_ = nonLinSolver;
    }

    /** \brief
     * Sets \ref updater.
     */
    inline void setUpdater(NonLinUpdaterVec *updater) {
      updater_ = updater;
    }


  protected:
    /** \brief
     * Initial guess for the solution.
     */
    Vector<AbstractFunction<double, WorldVector<double> >*> u0_;

    /** \brief
     * Non linear solver. Used in \ref solve().
     */
    NonLinSolver<SystemVector> *nonLinSolver_;

    /** \brief
     * Updator
     */
    NonLinUpdaterVec *updater_;
  };

}

#endif 

#endif

