#include "RCNeighbourList.h"
#include "Element.h"
#include "Line.h"
#include "Triangle.h"
#include "Tetrahedron.h"
#include "MacroElement.h"
#include "Mesh.h"
#include "Traverse.h"
#include "FixVec.h"
#include "CoarseningManager.h"
#include "PeriodicBC.h"
#include "DOFVector.h"
#include "LeafData.h"

namespace AMDiS {

  RCNeighbourList::RCNeighbourList(int maxEdgeNeigh) { 
    rclist.resize(maxEdgeNeigh); 
    for (int i = 0; i < maxEdgeNeigh; i++) {
      rclist[i] = new RCListElement;
    }
  }

  RCNeighbourList::~RCNeighbourList() {
    for (int i = 0; i < static_cast<int>(rclist.size()); i++) {
      delete rclist[i];
    }
  }

  /****************************************************************************/
  /*  do_coarse_patch:  if patch can be coarsend return true, else false and  */
  /*  reset the element marks                                                 */
  /****************************************************************************/

  bool RCNeighbourList::doCoarsePatch(int n_neigh)
  {
    FUNCNAME("RCNeighbourList::doCoarsePatch()");
  
    for (int i = 0; i < n_neigh; i++) {
      Element *lel = rclist[i]->el;

      if (lel->getMark() >= 0  ||  lel->isLeaf()) {
	/****************************************************************************/
	/*  element must not be coarsend or element is a leaf element, reset the    */
	/*  the coarsening flag on all those elements that have to be coarsend with */
	/*  this element                                                            */
	/****************************************************************************/
	lel->setMark(0);
	for (int j = 0; j < n_neigh; j++)
	  if (rclist[j]->flag)  
	    rclist[j]->el->setMark(0);

	return false;
      } else if (lel->getFirstChild()->getMark() >= 0  ||  
		 lel->getSecondChild()->getMark() >= 0) {

	/****************************************************************************/
	/*  one of the element's children must not be coarsend; reset the coarsening*/
	/*  flag on all those elements that have to be coarsend with this element   */
	/****************************************************************************/
	lel->setMark(0);
	for (int j = 0; j < n_neigh; j++)
	  if (rclist[j]->flag)  
	    rclist[j]->el->setMark(0);

	return false;
      } else if (!lel->getFirstChild()->isLeaf()  ||  
		 !lel->getSecondChild()->isLeaf())  {
	/****************************************************************************/
	/*  one of the element's children is not a leaf element;                    */
	/*  element may be coarsend after coarsening one of the children; try again */
	/****************************************************************************/
	coarseningManager->doMore = true;

	return false;
      } else {
	/****************************************************************************/ 
	/*  either one element is a macro element or we can coarsen the patch       */
	/****************************************************************************/
	if (rclist[i]->flag == 0) {

	  std::deque<MacroElement*>::const_iterator  mel;
	
	  Mesh* coarse_mesh = coarseningManager->getMesh(); 
	  // set in Mesh::coarsen()
	  for (mel = coarse_mesh->firstMacroElement(); 
	       mel!=coarse_mesh->endOfMacroElements(); 
	       mel++)
	    if ((*mel)->getElement() == lel)  
	      break;

	  TEST_EXIT(mel!=coarse_mesh->endOfMacroElements())
	    ("incompatible coarsening patch found\n");
	}
      }
    }

    return true;
  }


  void  RCNeighbourList::getNeighOnPatch(int n_neigh, int bound)
  {
    FUNCNAME("RCNeighbourList::getNeighOnPatch()");

    int j, k, dir;
    Element *el, *neigh;

    for (int i = 0; i < n_neigh; i++) {
      el = rclist[i]->el;
      rclist[i]->no = i;
      
      for (dir = 0; dir < 2; dir++) {
	for (j = 0; j < n_neigh; j++) {
	  if ((neigh = rclist[j]->el) == el)  
	    continue;
	  
	  for (k = 0; k < 2; k++) {
	    if (neigh->getDOF(2+k) == el->getDOF(3-dir)) {
	      rclist[i]->neigh[dir] = rclist[j];
	      rclist[i]->oppVertex[dir] = 3-k;
	      break;
	    }
	  }
	  
	  if (k < 2) 
	    break;
	}

	if (j >= n_neigh) {
	  rclist[i]->neigh[dir] = NULL;
	  rclist[i]->oppVertex[dir] = -1;
	}
      }
    }
  }

  void RCNeighbourList::addDOFParent(int elIndex, DegreeOfFreedom* dof)  // 3d
  {
    Element *el = rclist[elIndex]->el;
    RCListElement *neighbour;
    int        node;
  
    Mesh *coarse_mesh = coarseningManager->getMesh();
    RCListElement *coarse_list = rclist[elIndex];
  
    if (coarse_mesh->getNumberOfDOFs(EDGE))
      {
	node = coarse_mesh->getNode(EDGE);

	/****************************************************************************/
	/*  set the dof in the coarsening edge                                      */
	/****************************************************************************/

	el->setDOF(node, dof);

	/****************************************************************************/
	/*  and now those handed on by the children                                 */
	/****************************************************************************/

	el->setDOF(node+1, const_cast<int*>( el->getFirstChild()->getDOF(node)));
	el->setDOF(node+2, const_cast<int*>( el->getFirstChild()->getDOF(node+1)));
	el->setDOF(node+5, const_cast<int*>( el->getFirstChild()->getDOF(node+3)));
    
	if (coarse_list->elType)
	  {
	    el->setDOF(node+3, const_cast<int*>( el->getSecondChild()->getDOF(node)));
	    el->setDOF(node+4, const_cast<int*>( el->getSecondChild()->getDOF(node+1)));
	  }
	else
	  {
	    el->setDOF(node+3, const_cast<int*>( el->getSecondChild()->getDOF(node+1)));
	    el->setDOF(node+4, const_cast<int*>( el->getSecondChild()->getDOF(node)));
	  }
      }

    if (coarse_mesh->getNumberOfDOFs(FACE))
      {
	node = coarse_mesh->getNode(FACE);
	/****************************************************************************/
	/*  dof's at the faces within the patch: add new dof's if it is a boundary  */
	/*  face or neighbour is an element behind this element in the corsen list  */
	/****************************************************************************/
    
	neighbour = coarse_list->neigh[0];
	if ((!neighbour) || (neighbour > coarse_list))
	  {
	    if (!el->getDOF(node+2))
	      {
		el->setDOF(node+2, const_cast<int*>( coarse_mesh->getDOF(FACE)));   /* face  2 */
		if (neighbour)
		  neighbour->el->setDOF(node+coarse_list->oppVertex[0], 
					const_cast<int*>( el->getDOF(node+2)));
	      }
	  }
	neighbour = coarse_list->neigh[1];
	if ((!neighbour) || (neighbour > coarse_list))
	  {
	    if (!el->getDOF(node+3))
	      {
		el->setDOF(node+3, const_cast<int*>( coarse_mesh->getDOF(FACE)));    /* face  3 */
		if (neighbour)
		  neighbour->el->setDOF(node+coarse_list->oppVertex[1], 
					const_cast<int*>( el->getDOF(node+3)));
	      }
	  }
	/****************************************************************************/
	/*  and now those handed on by the children                                 */
	/****************************************************************************/

	el->setDOF(node, const_cast<int*>( el->getSecondChild()->getDOF(node+3)));
	el->setDOF(node+1, const_cast<int*>( el->getFirstChild()->getDOF(node+3)));
      }
  
    if (coarse_mesh->getNumberOfDOFs(CENTER))
      {
	node = coarse_mesh->getNode(CENTER);
	if (!el->getDOF(node))
	  el->setDOF(node, const_cast<int*>( coarse_mesh->getDOF(CENTER)));
      }
    return;  
  }

  void RCNeighbourList::addDOFParents(int n_neigh) // 2d 
  {
    int    i, node;

    Mesh *coarse_mesh = coarseningManager->getMesh();

    if (coarse_mesh->getNumberOfDOFs(EDGE))
      {
	node = coarse_mesh->getNode(EDGE);

	/****************************************************************************/
	/*  get dofs on the boundary of the coarsening patch from the children      */
	/****************************************************************************/
	for (i = 0; i < n_neigh; i++)
	  {
	    rclist[i]->el->
	      setDOF(node, const_cast<int*>( rclist[i]->el->getSecondChild()->getDOF(node+2)));
	    rclist[i]->el->
	      setDOF(node+1, const_cast<int*>( rclist[i]->el->getFirstChild()->getDOF(node+2)));
	  }
      }

    if (coarse_mesh->getNumberOfDOFs(CENTER))
      {
	int   node = coarse_mesh->getNode(CENTER);

	/****************************************************************************/
	/*  get new dof on parents at the barycenter                                */
	/****************************************************************************/
	for (i = 0; i < n_neigh; i++)
	  {
	    if (!rclist[i]->el->getDOF(node))
	      rclist[i]->el->setDOF(node, const_cast<int*>( coarse_mesh->getDOF(CENTER)));
	  }
      }

    return;
  }


  void RCNeighbourList::removeDOFParents(int n_neigh)
  {
    int    i, j, node;
    Mesh *mesh = rclist[0]->el->getMesh();
    int edges = mesh->getGeo(EDGE);

    if (mesh->getNumberOfDOFs(EDGE))
      {  
	node = mesh->getNode(EDGE);

	for (i = 0; i < n_neigh; i++)
	  {
	    for (j = 0; j < edges; j++)
	      rclist[i]->el->setDOF(node+j, NULL);
	  }
      }

    if (mesh->getNumberOfDOFs(CENTER))
      {
	node = mesh->getNode(CENTER);
	for (i = 0; i < n_neigh; i++)
	  {
	    mesh->freeDOF(const_cast<int*>( rclist[i]->el->getDOF(node)), CENTER);
	    rclist[i]->el->setDOF(node, NULL);
	  }
      }
  }

  void RCNeighbourList::removeDOFParent(int index)
  {
    Tetrahedron *el = dynamic_cast<Tetrahedron*>(rclist[index]->el);
    RCListElement *neigh;
    int        j, node;
    Mesh *mesh = el->getMesh();
    int edges = mesh->getGeo(EDGE);
    int faces = mesh->getGeo(FACE);


    if (mesh->getNumberOfDOFs(EDGE))
      {
	node = mesh->getNode(EDGE);
	for (j = 0; j < edges; j++)
	  el->setDOF(node+j, NULL);
      }

    if (mesh->getNumberOfDOFs(FACE))
      {
	node = mesh->getNode(FACE);
    
	neigh = rclist[index]->neigh[0];
	if ((!neigh) || (neigh > rclist[index]))
	  mesh->freeDOF(const_cast<int*>( el->getDOF(node+2)), FACE);           /* face  2 */

	neigh = rclist[index]->neigh[1];
	if ((!neigh) || (neigh > rclist[index]))
	  mesh->freeDOF(const_cast<int*>( el->getDOF(node+3)), FACE);           /* face  3 */

	for (j = 0; j < faces; j++)
	  el->setDOF(node+j, NULL);
      }
  
    if (mesh->getNumberOfDOFs(CENTER))
      {
	node = mesh->getNode(CENTER);
	mesh->freeDOF(const_cast<int*>( el->getDOF(node)), CENTER);
	el->setDOF(node, NULL);
      }
    return;
  }

  RCNeighbourList *RCNeighbourList::periodicSplit(DegreeOfFreedom *edge[2],
						  DegreeOfFreedom *nextEdge[2],
						  int *n_neigh,
						  int *n_neigh_periodic)
  {
    RCNeighbourList *periodicList = NULL;
    *n_neigh_periodic = 0;
    int count = 0;
    int n_neigh_old = *n_neigh;
    bool secondPart = false;

    nextEdge[0] = NULL;
    nextEdge[1] = NULL;

    std::vector<RCListElement*>::iterator it = rclist.begin();
    std::vector<RCListElement*>::iterator insertIt;

    while (count < n_neigh_old) {
      DegreeOfFreedom *dof0 = const_cast<DegreeOfFreedom*>((*it)->el->getDOF(0));
      DegreeOfFreedom *dof1 = const_cast<DegreeOfFreedom*>((*it)->el->getDOF(1));

      if (dof0 != edge[0] && dof0 != edge[1]) {
	secondPart = true;
	if (!nextEdge[0]) {
	  nextEdge[0] = dof0;
	  nextEdge[1] = dof1;
	}
	++it;
      } else {
	(*n_neigh)--;
	(*n_neigh_periodic)++;

	if (!periodicList) {
	  periodicList = new RCNeighbourList;
	  insertIt = periodicList->rclist.end();
	  periodicList->coarseningManager = coarseningManager;
	} else {
	  if (edge[0]) {
	    TEST_EXIT_DBG((dof0 == edge[0] && dof1 == edge[1]) ||
			  (dof1 == edge[0] && dof0 == edge[1]))
	      ("invalid macro file?\n");
	  }
	}

	if (secondPart) {
	  insertIt = periodicList->rclist.begin();
	  secondPart = false;
	} 

	insertIt = periodicList->rclist.insert(insertIt, *it);
	++insertIt;

	it = rclist.erase(it);
      } 
      count++;
    }

    for (int i = 0; i < *n_neigh_periodic; i++) {
      periodicList->rclist[i]->no = i;
    }

    return periodicList;
  }

}
