// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file Recovery.h */

#ifndef AMDIS_RECOVERY_H
#define AMDIS_RECOVERY_H

#include "set"
#include "Lagrange.h"
#include "Traverse.h"
#include "DOFVector.h"
#include "Cholesky.h"

using namespace std;
using namespace AMDiS;

class Monomial : public
BinaryAbstractFunction<double, WorldVector<double>, WorldVector<double> >
{
public:

  // Constructor.
  Monomial(WorldVector<int> expon)
    : BinaryAbstractFunction<double, WorldVector<double>, WorldVector<double> >(),
      exponent(expon)
  {
    degree_ = exponent[0];
    for (int i = 1; i<exponent.size(); i++)
      degree_ += exponent[i];
  }

  virtual ~Monomial() {};

  double operator()(const WorldVector<double>& y,
		    const WorldVector<double>& z) const
  {
    double result = pow(y[0] - z[0], exponent[0]);
    for (int i = 1; i < exponent.size(); i++)
      result *= pow(y[i] - z[i], exponent[i]);

    return result;
  }

private:
  WorldVector<int> exponent;
};


class RecoveryStructure
{
public:     // construtor, destructor
  // constructor
  RecoveryStructure() :
    coords(NULL),
    A(NULL), 
    rec_uh(NULL), 
    rec_grdUh(NULL),
    neighbors(NULL)
  {}

  // Destructor?
  ~RecoveryStructure()
  {
    clear();
  }

  RecoveryStructure(const RecoveryStructure& rhs) 
    : coords(NULL),
      A(NULL), 
      rec_uh(NULL), 
      rec_grdUh(NULL),
      neighbors(NULL) 
  {
    *this=rhs;
  }

  RecoveryStructure& operator=(const RecoveryStructure& rhs);


  // Clear recovery structure
  inline void clear()
  {
    if (coords)
      {
	delete coords;
	coords    = NULL;
      }

    if (A)
      {
	delete A;
	A         = NULL;
      }

    if (rec_uh)
      {
	delete rec_uh;
	rec_uh    = NULL;
      }

    if (rec_grdUh)
      {
	delete rec_grdUh;
	rec_grdUh = NULL;
      }

    if (neighbors)
      {
	delete neighbors;
	neighbors = NULL;
      }
  };

  // Prints recovery structure (for test purposes)
  void print();


private:     // members

  // World coordinates of the node.
  WorldVector<double> *coords;

  // For interior nodes.
  Matrix<double> *A;
  Vector<double> *rec_uh;
  Vector<WorldVector<double> > *rec_grdUh;

  // For boundary, edge, face, of center nodes: interior neighbors nodes.
  // For interior vertices in uh-recovery: neighbors nodes.
  std::set<DegreeOfFreedom> *neighbors;

  friend class Recovery;
};


// ============================================================================
// ===== class Recovery =======================================================
// ============================================================================

/**
 * \ingroup Recovery
 *
 * \brief
 * Recovering the gradient of a finite element function.
 */

class Recovery
{
public:
  // constructor
  Recovery(int norm, int method_)
    : struct_vec(NULL), feSpace(NULL), matrix_fcts(NULL), method(method_)
  {
    n_monomials = 0;
    gradient    = (norm == H1_NORM);
  };

  // Destructor?

  // Clear vector of recovery structures
  inline void clear()
  {
    if (struct_vec)
      {
	DOFVector<RecoveryStructure>::Iterator SV_it(struct_vec, ALL_DOFS);
	for (SV_it.reset(); !SV_it.end(); ++SV_it)
	  (*SV_it).clear();
      }
  };

  /** \brief
   * Recovers flux or gradient of given DOFVector.
   */
  DOFVector<WorldVector<double> >*
  recovery(DOFVector<double> *uh,
	   AbstractFunction<double, WorldVector<double> > *f_vec = NULL,
	   AbstractFunction<double, double> *f_scal = NULL,
	   DOFVector<double> *aux_vec = NULL);

  DOFVector<WorldVector<double> >*
  recovery(DOFVector<double> *uh, const FiniteElemSpace *fe_space,
	   AbstractFunction<double, WorldVector<double> > *f_vec = NULL,
	   AbstractFunction<double, double> *f_scal = NULL,
	   DOFVector<double> *aux_vec = NULL);

  /** \brief
   * Computes higher order approximation of given DOFVector.
   */
  void recoveryUh(DOFVector<double> *uh, DOFVector<double> &rec_vec);
  DOFVector<double>* recoveryUh(DOFVector<double> *uh,
				const FiniteElemSpace *fe_space);

  // for test purposes
  void test(DOFVector<double> *uh, const FiniteElemSpace *fe_space);


private:     // Private functions.

  // Defines a new finite element space if necessary.
  void set_feSpace(const FiniteElemSpace *fe_space);

  // Fills vector of exponents of the monomials.
  int set_exponents(int degree);

  // Fills vector of recovery structures.
  void fill_struct_vec(DOFVector<double> *uh,
		       AbstractFunction<double, WorldVector<double> > *f_vec = NULL,
		       AbstractFunction<double, double> *f = NULL,
		       DOFVector<double> *aux_vec = NULL);

  // Compute integrals defining matrix and vector on element
  // (continuous ZZ-recovery)
  void compute_integrals(DOFVector<double> *uh, ElInfo *elInfo,
			 RecoveryStructure *rec_struct,
			 AbstractFunction<double, WorldVector<double> > *f_vec = NULL,
			 AbstractFunction<double, double> *f_scal = NULL,
			 DOFVector<double> *aux_vec = NULL);

  // Compute integrals defining matrix and vector on element
  // (superconvergent patch recovery)
  void compute_interior_sums(DOFVector<double> *uh, ElInfo *elInfo,
			     RecoveryStructure *rec_struct, Quadrature *quad,
			     AbstractFunction<double, WorldVector<double> > *f_vec = NULL,
			     AbstractFunction<double, double> *f_scal = NULL,
			     DOFVector<double> *aux_vec = NULL);

  void compute_node_sums(DOFVector<double> *uh, ElInfo *elInfo,
			 RecoveryStructure *rec_struct, DimVec<int> preDOFs,
			 int n_vertices, int n_edges, int n_faces);

  void compute_sums_linear(DOFVector<double> *uh, ElInfo *elInfo,
			   RecoveryStructure *rec_struct,
			   int vertex, DimVec<int> preDOFs, int n_vertices);

private:     // Private members.

  // Structure storing needed information.
  DOFVector<RecoveryStructure> *struct_vec;

  const FiniteElemSpace *feSpace;     // Working finite element space.

  int                       n_monomials;     // Number of monomials.
  Vector<WorldVector<int> > exponents;       // Exponents of the monomials.
  Matrix<Monomial*>        *matrix_fcts;     // Functions for system matrix.

  bool gradient;     // True if gradient or flux should be recovered;
                     // false if seeking for higher order approximation of uh.

  int method;        // 0: superconvergent patch recovery (discrete ZZ)
                     // 1: local L2-averaging (continuous ZZ-recovery)
                     // 2: simple averaging
};

#endif
