#include "Estimator.h"
#include "Assembler.h"

template<typename T>
void RobinBC<T>::fillLocalBC(DOFVector<T>* vector, 
			     ElInfo*                   elInfo,
			     const DegreeOfFreedom*    dofIndices,
			     const BoundaryType*       localBound,
			     int                       nBasFcts)
{
  int dim = elInfo->getMesh()->getDim();
  int i;

  if(neumannOperators) {
    for(i=0; i < dim+1; i++) {
      if(elInfo->getBoundOfBoundary(i) == boundaryType) {
	vector->assemble(1.0, elInfo, localBound, (*neumannOperators)[i]);
      }
    }
  }
}

template<typename T>
void RobinBC<T>::fillLocalBC(DOFMatrix*             matrix,
			     ElInfo*                elInfo,
			     const DegreeOfFreedom* dofIndices,
			     const BoundaryType*    localBound,
			     int                    nBasFcts) 
{
  int dim = elInfo->getMesh()->getDim();
  int i;

  if(robinOperators) {
    for(i=0; i < dim+1; i++) {
      if(elInfo->getBoundOfBoundary(i) == boundaryType) {
	matrix->assemble(-1.0, elInfo, localBound, (*robinOperators)[i]);
      }
    }
  }
}
  
template<typename T>
double RobinBC<T>::boundResidual(ElInfo *elInfo, Estimator<T> *estimator)
{
  int dim = elInfo->getMesh()->getDim();
  int i, j, iq;
  DimVec<double>  lambda(dim, NO_INIT);
  double n_A_grdUh, val, f;
  WorldVector normal, grdUh, A_grdUh;

  const DimVec<WorldVector> &Lambda = elInfo->getGrdLambda();
  double det = elInfo->getDet();

  int numPoints;
  int dow = Global::getGeo(WORLD);

  bool neumannQuad = false;

  TEST_EXIT(neumannOperators || robinOperators)
    ("neither neumann nor robin operators set!\n");

  if(!robinOperators) 
    neumannQuad = true;
  else {
    if(neumannOperators) {
      if((*neumannOperators)[0]->getAssembler()->
	 getZeroOrderAssembler()->getQuadrature()->getNumPoints() > 
	 (*robinOperators)[0]->getAssembler()->
	 getZeroOrderAssembler()->getQuadrature()->getNumPoints()) 
	{
	  neumannQuad = true;
	}
    }
  }

  Quadrature *quadrature = NULL;

  int face;
  for(face = 0; face < dim+1; face++) {

    elInfo->getNormal(face, normal);

    quadrature = 
      neumannQuad ? 
      (*neumannOperators)[face]->getAssembler()->
      getZeroOrderAssembler()->getQuadrature() :
      (*robinOperators)[face]->getAssembler()->
      getZeroOrderAssembler()->getQuadrature();

    numPoints = quadrature->getNumPoints();

    if(elInfo->getBoundOfBoundary(face) == boundaryType) {

      (*neumannOperators)[face]->getAssembler()->getZeroOrderAssembler()->
	initElement(elInfo);

      val = 0.0;
      for (iq = 0; iq < numPoints; iq++) {
	lambda = quadrature->getLambda(iq);

	estimator->bas_fcts->evalGrdUh(lambda, Lambda, estimator->uhEl, &grdUh);

	if (estimator->isDiag) {
	  for (i = 0; i < dow; i++)
	    A_grdUh[i] = (*(estimator->A))[i][i]*grdUh[i];
	} else {
	  for (i = 0; i < dow; i++)
	    for (A_grdUh[i] = j = 0; j < dow; j++)
	      A_grdUh[i] += (*(estimator->A))[i][j]*grdUh[j];
	}

	f = 0.0;

	if(robinOperators) {
	  double uhAtQp = 0.0;
	  for(i=0; i < estimator->bas_fcts->getNumber(); i++)
	    uhAtQp += estimator->uhEl[i]*
	      (*(estimator->bas_fcts->getPhi(i)))(lambda);
	  (*robinOperators)[face]->evalZeroOrder(elInfo, iq, &f);
	  f *= uhAtQp;
	}

	if(neumannOperators)
	  (*neumannOperators)[face]->evalZeroOrder(elInfo, iq, &f);

	n_A_grdUh = (normal*A_grdUh) - f; 
	val += quadrature->getWeight(iq)*sqr(n_A_grdUh);
      }
    }
  }

  if (estimator->norm == NO_NORM || estimator->norm == L2_NORM)
    return(estimator->C1*estimator->h2_from_det(det, dim)*det*val);
  else
    return(estimator->C1*det*val);
}
